function detect_text(filename, extension)
%gathers data from the text detection process from text candidate regions
%and from connected components within those regions
%makes a file called <filename>_region_data.txt and another called
%<filename>_cc_data.txt containing candidate data

% go through the first set of motions to transform the image -
% up through step 3
orig = imread([filename extension]);
imwrite(orig, [filename '.tif'], 'tif', 'Compression', 'none');
[rows_image cols_image dont_care] = size(orig);
y = rgb2gray(orig);
overall_image_thresh = graythresh(y) * 255;
b = strel(ones(3,3));
ybl = (imclose(imopen(y, b), b) + imopen(imclose(y, b), b)) / 2;
es = imdilate(ybl, b) - imerode(ybl, b);
edge_threshold = graythresh(es) * 255;
e = (es >= edge_threshold);
ec = imerode(imdilate(e, b), b);
imwrite(ec, [filename 'tc.jpg'], 'jpg');

% do the labeling algorithm as given in step 4
[labeled_ec num_candidate_regions] = local_labeling(ec);

% do region analysis as given in step 5

%region/cc data used for clustering; cc_coordinates, whose lines
%correspond to lines in cc data, are used directly in finding
%the bounding box around all valid components
next_region_data_index = 1;
region_data = zeros(1, 17);
next_cc_coordinates_index = 1;
cc_coordinates = zeros(1, 7);
next_cc_data_index = 1;
cc_data = zeros(1, 18);

%"stranded" arrays carry data about regions that have been discarded
%for one reason or another but may be brought back in based on
%location near a valid connected component
next_stranded_coordinates_index = 1;
stranded_cc_coordinates = zeros(1, 7);
next_stranded_region_index = 1;
stranded_region_data = zeros(1, 17);
next_stranded_cc_data_index = 1;
stranded_cc_data = zeros(1, 18);

for region_num = 1:num_candidate_regions
[curr_region region_top_corner] = isolate_padded_region(y, labeled_ec, region_num);
region_size = size(curr_region);

% don't even bother with this region if it itself is too small
% (< 0.25% of image) or too big ( > 90% of image)
[rows_region cols_region] = size(curr_region);
if rows_region * cols_region > 0.9 * rows_image * cols_image || rows_region * cols_region < 0.0025 * rows_image * cols_image
str = ['region ' num2str(region_num) ' eliminated due to size']
continue
end

% record some preliminary information about the region
curr_region_data = zeros(1, 17, 'double');
curr_region_data(1) = double(region_num);
curr_region_data(2:3) = ((region_top_corner + region_size - 1) + region_top_corner) / 2;
curr_region_data(4:5) = region_size;
curr_region_data(6) = rows_region * cols_region;
curr_region_data(7) = double(rows_region * cols_region) / double(rows_image * cols_image);
curr_region_data(10) = mean(reshape(curr_region, 1, rows_region * cols_region));
curr_region_data(11) = graythresh(curr_region) * 255;

%write out this region in grayscale and edge grouped versions
imwrite(curr_region, ['region_' num2str(region_num) '.jpg'], 'jpg');
curr_tc_image = labeled_ec(region_top_corner(1):region_top_corner(1) + rows_region - 1, region_top_corner(2):region_top_corner(2) + cols_region - 1);
curr_tc_image = (curr_tc_image == region_num);
imwrite(curr_tc_image, ['region_' num2str(region_num) '_from_tc.jpg'], 'jpg');

%threshold and label the candidate region
%thresholding dependent on whether foreground text is light or dark
region_fg_light = tc_gradient_detection(curr_region, curr_tc_image, region_num, edge_threshold);
threshed_region = [];
if region_fg_light == 1
threshed_region = (curr_region >= curr_region_data(11));
else
threshed_region = (curr_region <= curr_region_data(11));
end
[labeled_region labeled_ccs] = local_labeling(threshed_region);

%record some more information about this region
curr_region_data(8) = labeled_ccs;
curr_region_data(12) = region_fg_light;
if region_fg_light == 1
fg_pixels = curr_region(find(curr_region >= curr_region_data(11)));
bg_pixels = curr_region(find(curr_region < curr_region_data(11)));
else
fg_pixels = curr_region(find(curr_region <= curr_region_data(11)));
bg_pixels = curr_region(find(curr_region > curr_region_data(11)));
end
curr_region_data(13) = mean(reshape(fg_pixels, 1, size(fg_pixels, 1) * size(fg_pixels, 2)));
curr_region_data(14) = std(double(reshape(fg_pixels, 1, size(fg_pixels, 1) * size(fg_pixels, 2))));
curr_region_data(15) = mean(reshape(bg_pixels, 1, size(bg_pixels, 1) * size(bg_pixels, 2)));
curr_region_data(16) = std(double(reshape(bg_pixels, 1, size(bg_pixels, 1) * size(bg_pixels, 2))));
curr_region_data(17) = (size(fg_pixels, 1) * size(fg_pixels, 2)) / (rows_region * cols_region);

valid_ccs = 0;
for cc_num = 1:labeled_ccs

%filter out the components that are too small or not text-like
[curr_cc cc_top_corner] = isolate_region(labeled_region, labeled_region, cc_num);

cc_size = size(curr_cc);
rows_cc = cc_size(1);
cols_cc = cc_size(2);
gray_cc = y(region_top_corner(1) + cc_top_corner(1) - 1:region_top_corner(1) + cc_top_corner(1) + rows_cc - 2, region_top_corner(2) + cc_top_corner(2) - 1:region_top_corner(2) + cc_top_corner(2) + cols_cc - 2);

imwrite(curr_cc, ['region_' num2str(region_num) '_cc_' num2str(cc_num) '.jpg'], 'jpg');
%imwrite(gray_cc, ['region_' num2str(region_num) '_cc_' num2str(cc_num) '_gray.jpg'], 'jpg');

%filter improper width
if rows_cc <= 3
threshed_region = filter_out_component(threshed_region, labeled_region, cc_num);
str = ['cc ' num2str(cc_num) ' filtered due to improper width']
continue
end

%filter improper length
if cols_cc <= 3 
threshed_region = filter_out_component(threshed_region, labeled_region, cc_num);
str = ['cc ' num2str(cc_num) ' filtered due to improper length']
continue
end

%filter improper size ratio - see (8) in Hasan/Karam
ratios_cc = [rows_cc / cols_cc cols_cc / rows_cc];
[area_cc dont_care] = size(find(curr_cc == cc_num));
area_ratio = area_cc / (rows_cc * cols_cc);
if area_ratio <= 1 / 4.5 || area_ratio > 0.95
if min(ratios_cc) <= 1.5
threshed_region = filter_out_component(threshed_region, labeled_region, cc_num);
str = ['cc ' num2str(cc_num) ' filtered due to improper size ratio']
continue
end
end

%now it's an okay component - save some information about it
curr_cc_data = zeros(1, 18, 'double');
curr_cc_data(1) = region_num;
curr_cc_data(2) = cc_num;
curr_cc_data(3:4) = ((region_top_corner + cc_top_corner - 1) + (region_top_corner + cc_top_corner + cc_size - 2)) / 2;
curr_cc_data(5) = rows_cc;
curr_cc_data(6) = cols_cc;
curr_cc_data(7) = rows_cc * cols_cc;
curr_cc_data(8) = curr_cc_data(7) / curr_region_data(6);
curr_cc_data(9) = area_ratio;
curr_cc_data(10) = mean(reshape(gray_cc, 1, rows_cc * cols_cc));
curr_cc_data(11) = graythresh(gray_cc) * 255;

[curr_padded_cc padded_cc_top_corner] = isolate_padded_region(y, labeled_region, cc_num);
[padded_rows_cc padded_cols_cc] = size(curr_padded_cc);
curr_cc_tc_image = curr_tc_image(padded_cc_top_corner(1):padded_cc_top_corner(1) + padded_rows_cc - 1, padded_cc_top_corner(2):padded_cc_top_corner(2) + padded_cols_cc - 1);

cc_fg_light = tc_gradient_detection(curr_padded_cc, curr_cc_tc_image, region_num, edge_threshold);
curr_cc_data(12) = cc_fg_light;
if cc_fg_light == 1
fg_pixels = gray_cc(find(gray_cc >= curr_cc_data(11)));
bg_pixels = gray_cc(find(gray_cc < curr_cc_data(11)));
else
fg_pixels = gray_cc(find(gray_cc <= curr_cc_data(11)));
bg_pixels = gray_cc(find(gray_cc > curr_cc_data(11)));
end
if size(fg_pixels, 1) ~= 0 && size(fg_pixels, 2) ~= 0
curr_cc_data(13) = mean(reshape(fg_pixels, 1, size(fg_pixels, 1) * size(fg_pixels, 2)));
curr_cc_data(14) = std(double(reshape(fg_pixels, 1, size(fg_pixels, 1) * size(fg_pixels, 2))));
end
if size(bg_pixels, 1) ~= 0 && size(bg_pixels, 2) ~= 0
curr_cc_data(15) = mean(reshape(bg_pixels, 1, size(bg_pixels, 1) * size(bg_pixels, 2)));
curr_cc_data(16) = std(double(reshape(bg_pixels, 1, size(bg_pixels, 1) * size(bg_pixels, 2))));
end
curr_cc_data(17) = (size(fg_pixels, 1) * size(fg_pixels, 2)) / (rows_cc * cols_cc);

cc_edge_image = es(region_top_corner(1) + padded_cc_top_corner(1) - 1:region_top_corner(1) + padded_cc_top_corner(1) + padded_rows_cc - 2, region_top_corner(2) + padded_cc_top_corner(2) - 1:region_top_corner(2) + padded_cc_top_corner(2) + padded_cols_cc - 2);
%imwrite(cc_edge_image, ['region_' num2str(region_num) '_cc_' num2str(cc_num) '_edgeimage.jpg'], 'jpg');
cc_edge_high_contrasts = cc_edge_image(find(cc_edge_image >= edge_threshold));
if size(cc_edge_high_contrasts, 1) ~= 0 && size(cc_edge_high_contrasts, 2) ~= 0
curr_cc_data(18) = mean(reshape(cc_edge_high_contrasts, 1, size(cc_edge_high_contrasts, 1) * size(cc_edge_high_contrasts, 2)));
end

%if it's an okay component, mark it valid and add it to cc_coordinates
valid_ccs = valid_ccs + 1;
str = ['keeping region ' num2str(region_num) ' cc ' num2str(cc_num)]

cc_coordinates(next_cc_coordinates_index, :) = [region_top_corner + cc_top_corner - 1, region_top_corner + cc_top_corner + cc_size - 2, region_num, cc_num, region_fg_light];
next_cc_coordinates_index = next_cc_coordinates_index + 1;
cc_data(next_cc_data_index, 1:size(curr_cc_data, 2)) = curr_cc_data;
next_cc_data_index = next_cc_data_index + 1;

%end loop over components w/in region
end

%keep track of last bit of region information
curr_region_data(9) = valid_ccs / labeled_ccs;

%if no or one connected components remain, this region is not text-like
if valid_ccs < 2
str = ['region ' num2str(region_num) ' contains ' num2str(valid_ccs) ' valid connected components']

%have to get rid of all cc coordinates for this region - but keep
%the coordinates in the "stranded" array for later analysis
if valid_ccs ~= 0

stranded_lines = find(cc_coordinates(:, 5) == region_num);
for stranded_index = 1:size(stranded_lines, 1) * size(stranded_lines, 2)
stranded_cc_coordinates(next_stranded_coordinates_index, 1:size(cc_coordinates, 2)) = cc_coordinates(stranded_lines(stranded_index), :);
next_stranded_coordinates_index = next_stranded_coordinates_index + 1;
stranded_cc_data(next_stranded_cc_data_index, 1:size(cc_data, 2)) = cc_data(stranded_lines(stranded_index), :);
next_stranded_cc_data_index = next_stranded_cc_data_index + 1;

found_region_line = find(stranded_region_data(:, 1) == cc_coordinates(stranded_lines(stranded_index), 5));
if size(found_region_line, 1) == 0 || size(found_region_line, 2) == 0
stranded_region_data(next_stranded_region_index, 1:size(region_data, 2)) = curr_region_data;
next_stranded_region_index = next_stranded_region_index + 1;
end

end

cc_coordinates = filter_cc_coordinates(cc_coordinates, region_num, -1);
next_cc_coordinates_index = size(cc_coordinates, 1) + 1;
cc_data = filter_cc_data(cc_data, region_num, -1);
next_cc_data_index = size(cc_data, 1) + 1;

end

%continue to analyze the next region
continue
end

str = ['region ' num2str(region_num) ' contains ' num2str(valid_ccs) ' valid connected components of ' num2str(labeled_ccs) ' identified components']

%erode once - see if it still contains any text after that
eroded_region = imerode(threshed_region , b);
[num_text_pixels dont_care] = size(find(eroded_region == 1));
if num_text_pixels == 0
str = ['region ' num2str(region_num) ' has no text pixels after erosion']

stranded_lines = find(cc_coordinates(:, 5) == region_num);
for stranded_index = 1:size(stranded_lines, 1) * size(stranded_lines, 2)
stranded_cc_coordinates(next_stranded_coordinates_index, 1:size(cc_coordinates, 2)) = cc_coordinates(stranded_lines(stranded_index), :);
next_stranded_coordinates_index = next_stranded_coordinates_index + 1;
stranded_cc_data(next_stranded_cc_data_index, 1:size(cc_data, 2)) = cc_data(stranded_lines(stranded_index), :);
next_stranded_cc_data_index = next_stranded_cc_data_index + 1;

found_region_line = find(stranded_region_data(:, 1) == cc_coordinates(stranded_lines(stranded_index), 5));
if size(found_region_line, 1) == 0 || size(found_region_line, 2) == 0
stranded_region_data(next_stranded_region_index, 1:size(region_data, 2)) = curr_region_data;
next_stranded_region_index = next_stranded_region_index + 1;
end

end

cc_coordinates = filter_cc_coordinates(cc_coordinates, region_num, -1);
next_cc_coordinates_index = size(cc_coordinates, 1) + 1;
cc_data = filter_cc_data(cc_data, region_num, -1);
next_cc_data_index = size(cc_data, 1) + 1;

%continue to analyze the next region
continue
end 

%if the region has survived with some components still marked,
%ensure the connected components (character-like objects) have
%sufficient gray-level contrast with background
in_cc = curr_region(labeled_region > 0);
[in_size dont_care] = size(in_cc);
non_cc = curr_region(labeled_region == 0);
[non_size dont_care] = size(non_cc);
avg_cc_level = sum(in_cc) / in_size;
avg_noncc_level = sum(non_cc) / non_size;
if abs(avg_cc_level - avg_noncc_level) < 30

str = ['region ' num2str(region_num) ' has invalid grey-level contrast.']
stranded_lines = find(cc_coordinates(:, 5) == region_num);

for stranded_index = 1:size(stranded_lines, 1) * size(stranded_lines, 2)
stranded_cc_coordinates(next_stranded_coordinates_index, 1:size(cc_coordinates, 2)) = cc_coordinates(stranded_lines(stranded_index), :);
next_stranded_coordinates_index = next_stranded_coordinates_index + 1;
stranded_cc_data(next_stranded_cc_data_index, 1:size(cc_data, 2)) = cc_data(stranded_lines(stranded_index), :);
next_stranded_cc_data_index = next_stranded_cc_data_index + 1;

found_region_line = find(stranded_region_data(:, 1) == cc_coordinates(stranded_lines(stranded_index), 5));
if size(found_region_line, 1) == 0 || size(found_region_line, 2) == 0
stranded_region_data(next_stranded_region_index, 1:size(region_data, 2)) = curr_region_data;
next_stranded_region_index = next_stranded_region_index + 1;
end

end

cc_coordinates = filter_cc_coordinates(cc_coordinates, region_num, -1);
next_cc_coordinates_index = size(cc_coordinates, 1) + 1;
cc_data = filter_cc_data(cc_data, region_num, -1);
next_cc_data_index = size(cc_data, 1) + 1;

%continue to analyze the next region
continue
end

imwrite(threshed_region, ['thregion_' num2str(region_num) '.jpg'], 'jpg');

%if the region has still survived, it's valid
%and we have kept all the connected components in it that we care about

%keep the region information in the region data matrix
region_data(next_region_data_index, 1:size(curr_region_data, 2)) = curr_region_data;
next_region_data_index = next_region_data_index + 1;

%end loop over regions
end

%remove all duplicates from the connected component data
[cc_coordinates cc_data] = remove_duplicate_data(cc_coordinates, cc_data, region_data)

%print the data and coordinates we do have
str = 'data before clustering'
region_data
cc_data
cc_coordinates
stranded_region_data
stranded_cc_data
stranded_cc_coordinates

str = 'about to call Weka...'

%file output! chyeah!
%this is where we write the files for region data and cc data
region_file = fopen([filename '_region_data.txt'], 'w');
cc_file = fopen([filename '_cc_data.txt'], 'w');

for region_index = 1:size(region_data, 1)
fprintf(region_file, '%g,%g,%g,%g,%g,%g,%g,%g,%g,%g,%g,%g,%g,%g,%g,%g,%g\n', region_data(region_index, 1), region_data(region_index, 2), region_data(region_index, 3), region_data(region_index, 4), region_data(region_index, 5), region_data(region_index, 6), region_data(region_index, 7), region_data(region_index, 8), region_data(region_index, 9), region_data(region_index, 10), region_data(region_index, 11), region_data(region_index, 12), region_data(region_index, 13), region_data(region_index, 14), region_data(region_index, 15), region_data(region_index, 16), region_data(region_index, 17));
end
for cc_index = 1:size(cc_data, 1)
fprintf(cc_file, '%g,%g,%g,%g,%g,%g,%g,%g,%g,%g,%g,%g,%g,%g,%g,%g,%g,%g\n', cc_data(cc_index, 1), cc_data(cc_index, 2), cc_data(cc_index, 3), cc_data(cc_index, 4), cc_data(cc_index, 5), cc_data(cc_index, 6), cc_data(cc_index, 7), cc_data(cc_index, 8), cc_data(cc_index, 9), cc_data(cc_index, 10), cc_data(cc_index, 11), cc_data(cc_index, 12), cc_data(cc_index, 13), cc_data(cc_index, 14), cc_data(cc_index, 15), cc_data(cc_index, 16), cc_data(cc_index, 17), cc_data(cc_index, 18));
end

fclose(region_file);
fclose(cc_file);

% run the bash script for getting clustering info out of Weka
command = ['./weka_clustering ' filename];
unix(command);

% read in all the clustering information for regions
region_cluster_file = fopen([filename '_region_data_clusters.txt']);
for region_index = 1:size(region_data,1)
curr_char = fscanf(region_cluster_file, '%c\n', 1);
if size(curr_char, 1) == 0 || size(curr_char, 2) == 0
str = 'something went wrong - not enough clustering data!'
return
end
region_data(region_index, 18) = str2num(curr_char);
end
fclose(region_cluster_file);

% read in all the clustering information for ccs
cc_cluster_file = fopen([filename '_cc_data_clusters.txt']);
for cc_index = 1:size(cc_data,1)
curr_char = fscanf(cc_cluster_file, '%c\n', 1);
if size(curr_char, 1) == 0 || size(curr_char, 2) == 0
str = 'something went wrong - not enough clustering data!'
return
end
cc_data(cc_index, 19) = str2num(curr_char);
end
fclose(cc_cluster_file);

%figure out size of all components
cc_lengths = cc_coordinates(:, 4) - cc_coordinates(:,2);
cc_widths = cc_coordinates(:, 3) - cc_coordinates(:,1);
cc_areas = cc_lengths .* cc_widths

%figure out the mean size of connected components and the region cluster
%that most of the CCs within 1/4 SD of the mean fall into
my_mean = mean(cc_areas);
my_std = std(double(cc_areas));
my_median = median(cc_areas);

candidate_indices = find((cc_areas > my_mean - 0.25 * my_std) & (cc_areas < my_mean + 0.25 * my_std));
candidate_ccs = cc_data(candidate_indices, :)
cluster_counts = zeros(2, 1);

for index = 1:size(candidate_ccs, 1)
candidate_region = candidate_ccs(index, 1);
candidate_cluster = region_data(find(region_data(:, 1) == candidate_region), 18);
cluster_counts(candidate_cluster) = cluster_counts(candidate_cluster) + 1;
end

if cluster_counts(1) > cluster_counts(2)
selected_cluster = 1;
else
selected_cluster = 2;
end

region_data

%filter out all connected components from the unselected region cluster
new_index = 1;
new_coordinates = [];
new_data = [];

for old_index = 1:size(cc_coordinates, 1)
region_line = find(region_data(:,1) == cc_data(old_index, 1));
curr_cluster = region_data(region_line, 18);

if curr_cluster == selected_cluster
new_coordinates(new_index, 1:size(cc_coordinates, 2)) = cc_coordinates(old_index, 1:size(cc_coordinates,2));
new_data(new_index, 1:size(cc_data, 2)) = cc_data(old_index, 1:size(cc_data,2));
new_index = new_index + 1;

%here's the filtering part, but keep it in the stranded bin so we may
%add it in later due to proximity and similarity
else
stranded_cc_coordinates(next_stranded_coordinates_index, 1:size(cc_coordinates, 2)) = cc_coordinates(old_index, 1:size(cc_coordinates,2));
next_stranded_coordinates_index = next_stranded_coordinates_index + 1;
stranded_cc_data(next_stranded_cc_data_index, 1:size(cc_data, 2)) = cc_data(old_index, 1:size(cc_data, 2));
next_stranded_cc_data_index = next_stranded_cc_data_index + 1;

found_region_line = find(stranded_region_data(:, 1) == cc_coordinates(old_index, 5));
if size(found_region_line, 1) == 0 || size(found_region_line, 2) == 0
stranded_region_data(next_stranded_region_index, 1:size(region_data, 2)) = region_data(region_line, :);
next_stranded_region_index = next_stranded_region_index + 1;

end

end

end

cc_coordinates = new_coordinates
cc_data = new_data

%figure out what relative intensity (light or dark) the majority
%of the saved connected components have
fg_light_components = size(find(cc_coordinates(:, 7) == 1), 1)
fg_dark_components = size(find(cc_coordinates(:, 7) == 0), 1)
if fg_light_components > fg_dark_components
final_fg_light = 1;
else
final_fg_light = 0;
end

%save only the cc coordinates with that intensity
new_coordinates = [];
new_data = [];
new_index = 1;
for old_index = 1:size(cc_coordinates, 1)
if cc_coordinates(old_index, 7) == final_fg_light
new_coordinates(new_index, 1:size(cc_coordinates, 2)) = cc_coordinates(old_index, 1:size(cc_coordinates, 2));
new_data(new_index, 1:size(cc_data, 2)) = cc_data(old_index, 1:size(cc_data, 2));
new_index = new_index + 1;
end
end

cc_coordinates = new_coordinates
cc_data = new_data

%clean up region data - only have regions in it that are represented
%by at least one component
region_index = 1;
while region_index <= size(region_data, 1)
representative_ccs = find(cc_data(:, 1) == region_data(region_index, 1));
if size(representative_ccs, 1) == 0 || size(representative_ccs, 2) == 0
region_data(region_index, :) = [];
else
region_index = region_index + 1;
end
end

region_data

%pull components discarded b/c of region analysis or clustering back in
%if they are close enough to existing words 
[new_cc_coordinates new_cc_data new_region_data] = pull_in_stranded_ccs(cc_coordinates, cc_data, region_data, stranded_cc_coordinates, stranded_cc_data, stranded_region_data);

cc_coordinates = new_cc_coordinates
cc_data = new_cc_data
region_data = new_region_data

%filter out all connected components from the unselected region cluster
%that may have crept in when we were pulling in
new_index = 1;
new_coordinates = [];
new_data = [];

for old_index = 1:size(cc_coordinates, 1)
region_line = find(region_data(:,1) == cc_data(old_index, 1));
curr_cluster = region_data(region_line, 18)

if curr_cluster == selected_cluster || round(curr_cluster) == 0
new_coordinates(new_index, 1:size(cc_coordinates, 2)) = cc_coordinates(old_index, 1:size(cc_coordinates,2));
new_data(new_index, 1:size(cc_data, 2)) = cc_data(old_index, 1:size(cc_data,2));
new_index = new_index + 1;
end

end

cc_coordinates = new_coordinates
cc_data = new_data

%clean up region data again
region_index = 1;
while region_index <= size(region_data, 1)
representative_ccs = find(cc_data(:, 1) == region_data(region_index, 1));
if size(representative_ccs, 1) == 0 || size(representative_ccs, 2) == 0
region_data(region_index, :) = [];
else
region_index = region_index + 1;
end
end

region_data 

% -- after all that clustering mess --
% find final bounding box
min_row = min(cc_coordinates(:, 1));
min_col = min(cc_coordinates(:, 2));
max_row = max(cc_coordinates(:, 3));
max_col = max(cc_coordinates(:, 4));

%adjust final bounding box so we aren't so severe
left_wiggle = min_col;
right_wiggle = cols_image - max_col;
up_wiggle = min_row;
down_wiggle = rows_image - max_row;

box_top = max(min_row - 5, min_row - up_wiggle + 1)
box_bottom = min(max_row + 5, max_row + down_wiggle - 1)
box_left = max(min_col - 5, min_col - left_wiggle + 1)
box_right = min(max_col + 5, max_col + right_wiggle - 1)

%find and write out final region image
final_region_image = zeros(box_bottom - box_top + 1, box_right - box_left + 1);
[num_regions dont_care] = size(cc_coordinates);

included_regions = unique(cc_coordinates(:,5));
for curr_row = box_top:box_bottom
for curr_col = box_left:box_right
curr_region_num = labeled_ec(curr_row, curr_col);
if size(find(included_regions == curr_region_num), 1) ~= 0 && size(find(included_regions == curr_region_num), 2) ~= 0 
final_region_image(curr_row - box_top + 1, curr_col - box_left + 1) = curr_region_num;
end
end
end

save([filename '_processed_all_regions.mat'], 'final_region_image');
imwrite(final_region_image, [filename '_processed_all_regions.tif'], 'tif');

%write final text image

final_image = orig(box_top:box_bottom, box_left:box_right, :);
final_gray = rgb2gray(final_image);
imwrite(final_image, [filename '_final_orig.jpg'], 'jpg');

%convert to black and white - thresholding is backward
%from usual because we want final image to have black text and
%white background

if final_fg_light == 1
threshed_final = (final_gray <= graythresh(final_gray) * 255);
else
threshed_final = (final_gray >= graythresh(final_gray) * 255);
end

imwrite(threshed_final, [filename '_processed.tif'], 'tif', 'Compression', 'none');


%% end of main %%
%% start of local functions %%

function [labeled_image num_regions] = local_labeling(orig_img)
% find all the connected components in this region using local
% labeling! algorithm given in Hasan/Karam paper in pseudocode, step 4
% N = number of rows, M = number of columns
[N, M] = size(orig_img);
labeled_img = zeros(N,M);

%assign temporary labels to nonzero pixels of the image
current_label = 1;
for i = 1:N
for j = 1:M
if orig_img(i,j) == 1
labeled_img(i,j) = current_label;
current_label = current_label + 1;
end
end
end
max_temp_label = current_label - 1; %cause we always add 1 after assigning

%propagate maximum label per connected component
changed = 1;

%loop while at least one pixel value of labeled_ec changes
while changed == 1

%clear out last iteration's changes
changed = 0;   
  
%back propagation row-wise
for i=N:2
for j=1:M
if labeled_img(i, j) > 0 && labeled_img(i-1, j) > 0
old_first = labeled_img(i, j);
old_second = labeled_img(i-1, j);
if old_first ~= old_second
     changed = 1;
end
max_old = max(old_first, old_second);
labeled_img(i, j) = max_old;
labeled_img(i-1, j) = max_old;
end
end
end

%back propagation column wise
for i=1:N
for j=M:2
if labeled_img(i, j) > 0 && labeled_img(i, j-1) > 0
old_first = labeled_img(i, j);
old_second = labeled_img(i, j-1);
if old_first ~= old_second
     changed = 1;
end
max_old = max(old_first, old_second);
labeled_img(i, j) = max_old;
labeled_img(i, j-1) = max_old;
end
end
end

%forward propagation row-wise
for i=1:N-1
for j=1:M
if labeled_img(i, j) > 0 && labeled_img(i+1, j) > 0
old_first = labeled_img(i, j);
old_second = labeled_img(i+1, j);
if old_first ~= old_second
     changed = 1;
end
max_old = max(old_first, old_second);
labeled_img(i, j) = max_old;
labeled_img(i+1, j) = max_old;
end
end
end

%forward propagation column wise
for i=1:N
for j=1:M-1
if labeled_img(i, j) > 0 && labeled_img(i, j+1) > 0
old_first = labeled_img(i, j);
old_second = labeled_img(i, j+1);
if old_first ~= old_second
     changed = 1;
end
max_old = max(old_first, old_second);
labeled_img(i, j) = max_old;
labeled_img(i, j+1) = max_old;
end
end
end

%finished propagation
end

%sequential relabeling
array_label = zeros(1, max_temp_label);
max_label = 0;
for i = 1:N
for j = 1:M
if orig_img(i,j) == 1
if array_label(labeled_img(i,j)) == 0
max_label = max_label + 1;
array_label(labeled_img(i,j)) = max_label;
labeled_img(i,j) = max_label;
else
labeled_img(i,j) = array_label(labeled_img(i,j));
end
end
end
end

num_regions = max_label;
labeled_image = labeled_img;

function filtered_img = filter_out_component(img_to_filter, labeled_img, cc_num)
%filter out all pixels marked as part of the component numbered cc_num
%(set them to black so that component is now not a part of the region)
img_to_filter(labeled_img == cc_num) = 0;
filtered_img = img_to_filter;

function new_coordinates = filter_cc_coordinates(cc_coordinates, region_num, cc_num)
%gilter out the coordinates from the region and perhaps cc number within that
%region that we no longer care about
%if the provided cc_num is -1, filter out all ccs from the given region_num

new_coordinates = [];
[rows_coord cols_coord] = size(cc_coordinates);

%find the rows in cc_coordinates that we want to take out
if cc_num == -1
candidate_rows = find(cc_coordinates(:,5) == region_num);
else
candidate_rows = find(cc_coordinates(:,5) == region_num & cc_coordinates(:,6) == cc_num);
end

%copy over only the rows we like
new_index = 1;
for index = 1:size(cc_coordinates, 1)
if size(find(candidate_rows == index), 1) == 0 || size(find(candidate_rows == index), 2) == 0
new_coordinates(new_index, 1:cols_coord) = cc_coordinates(index, :);
new_index = new_index + 1;
end
end

function new_data = filter_cc_data(cc_data, region_num, cc_num)
%filter out the data from the region and perhaps cc number within that
%region that we no longer care about
%if the provided cc_num is -1, filter out all ccs from the given region_num

new_data = [];
[rows_data cols_data] = size(cc_data);

%find the rows in cc_data that we want to take out
if cc_num == -1
candidate_rows = find(cc_data(:,1) == region_num);
else
candidate_rows = find(cc_data(:,1) == region_num & cc_data(:,2) == cc_num);
end

%copy over only the rows we like
new_index = 1;
for index = 1:size(cc_data, 1)
if size(find(candidate_rows == index), 1) == 0 || size(find(candidate_rows == index), 2) == 0
new_data(new_index, 1:cols_data) = cc_data(index, :);
new_index = new_index + 1;
end
end

function [region_image top_corner] = isolate_region(y, labeled_img, region_num)
% isolates the region of the image labeled
% parameter y is the array from which the isolation should be made
% (corresponds to y in top level of algorithm for region selection before
% region analysis, but corresponds to thresholded region when testing
% for non-character components in candidate regions)
% give up to 2 px padding in all directions, if possible
[N, M] = size(labeled_img);
min_row = -1;
min_col = -1;
max_row = -1;
max_col = -1;

for i = 1:N

for j = 1:M

if labeled_img(i,j) == region_num

if min_row == -1 || i < min_row
min_row = i;
end

if min_col == -1 || j < min_col
min_col = j;
end

if max_row == -1 || i > max_row
max_row = i;
end

if max_col == -1 || j > max_col
max_col = j;
end

end

end
end

%slice out the grayscale region containing the
%connected component with this number
region_image = y(min_row:max_row, min_col:max_col);
top_corner = [min_row, min_col];

function [region_image top_corner] = isolate_padded_region(y, labeled_img, region_num)
% isolates the region of the image labeled
% parameter y is the array from which the isolation should be made
% (corresponds to y in top level of algorithm for region selection before
% region analysis, but corresponds to thresholded region when testing
% for non-character components in candidate regions)
% give up to 2 px padding in all directions, if possible
[N, M] = size(labeled_img);
min_row = -1;
min_col = -1;
max_row = -1;
max_col = -1;

for i = 1:N

for j = 1:M

if labeled_img(i,j) == region_num

if min_row == -1 || i < min_row
min_row = i;
end

if min_col == -1 || j < min_col
min_col = j;
end

if max_row == -1 || i > max_row
max_row = i;
end

if max_col == -1 || j > max_col
max_col = j;
end

end

end
end

% add padding if possible
min_row = max(1, min_row - 2);
max_row = min(N, max_row + 2);
min_col = max(1, min_col - 2);
max_col = min(M, max_col + 2);

%slice out the grayscale region containing the
%connected component with this number
region_image = y(min_row:max_row, min_col:max_col);
top_corner = [min_row, min_col];


function is_fg_light = tc_gradient_detection(curr_region, curr_tc_image, region_num, edge_threshold)

[rows_region cols_region] = size(curr_region);

% find the average gradient (up, or fg light, is 1; down, or fg dark, is -1;
% neither is 0) in each of four directions (up 1, down 2, left 3, right 4)
average_directional_gradient = zeros(4, 1);
for grad_direction = 1:4

% set up vector of gradient information for this direction
if grad_direction <= 2
num_samples = size(curr_region, 2);
else
num_samples = size(curr_region, 1);
end
grad_directional_gradient = zeros(num_samples, 1);

% find gradient information along each row (l/r) or column (u/d)
% for this direction
for curr_sample = 1:num_samples

%find first TC pixel along the line in this direction
if grad_direction == 1
possible_pixels = find(curr_tc_image(:, curr_sample) == 1);
if size(possible_pixels, 1) ~= 0 && size(possible_pixels, 2) ~= 0
first_pixel_index = max(possible_pixels);
else
continue
end
end
if grad_direction == 2
possible_pixels = find(curr_tc_image(:, curr_sample) == 1);
if size(possible_pixels, 1) ~= 0 && size(possible_pixels, 2) ~= 0
first_pixel_index = min(possible_pixels);
else
continue
end
end
if grad_direction == 3
possible_pixels = find(curr_tc_image(curr_sample, :) == 1);
if size(possible_pixels, 1) ~= 0 && size(possible_pixels, 2) ~= 0
first_pixel_index = min(possible_pixels);
else
continue
end
end
if grad_direction == 4
possible_pixels = find(curr_tc_image(curr_sample, :) == 1);
if size(possible_pixels, 1) ~= 0 && size(possible_pixels, 2) ~= 0
first_pixel_index = max(possible_pixels);
else
continue
end
end

% if the pixel's on the image boundary, continue to next sample
if (grad_direction <= 2 && (first_pixel_index ==1 || first_pixel_index == rows_region)) || (grad_direction > 2 && (first_pixel_index ==1 || first_pixel_index == cols_region))
continue
end

% find average gray value of pixels leading up to that fixst index along
% the line in this direction
if grad_direction == 1
pre_pixels = curr_region(first_pixel_index + 1:rows_region,curr_sample);
end
if grad_direction == 2
pre_pixels = curr_region(1:first_pixel_index - 1,curr_sample);
end
if grad_direction == 3
pre_pixels = curr_region(curr_sample, 1:first_pixel_index - 1);
end
if grad_direction == 4
pre_pixels = curr_region(curr_sample, first_pixel_index + 1:cols_region);
end
avg_pre_gray_level = mean(pre_pixels);

% look along the TC edge pixels on this line until we get a drastic
% (ie. above threshold for es) gray-level difference 
curr_index = first_pixel_index;

while (grad_direction == 1 && curr_index >= 1 && curr_tc_image(curr_index, curr_sample) == 1) || (grad_direction == 2 && curr_index <= rows_region && curr_tc_image(curr_index, curr_sample) == 1) || (grad_direction == 3 && curr_index <= cols_region && curr_tc_image(curr_sample, curr_index) == 1) || (grad_direction == 4 && curr_index >= 1 && curr_tc_image(curr_sample, curr_index) == 1)

if grad_direction <= 2
curr_difference = double(curr_region(curr_index, curr_sample)) - avg_pre_gray_level;
end
if grad_direction > 2
curr_difference = double(curr_region(curr_sample, curr_index)) - avg_pre_gray_level;
end

% save gradient information if
if abs(curr_difference) > edge_threshold
if curr_difference > 0
grad_directional_gradient(curr_sample) = 1;
else
grad_directional_gradient(curr_sample) = -1;
end
break
end

if grad_direction == 2 || grad_direction == 3
curr_index = curr_index + 1;
else
curr_index = curr_index - 1;
end

% end look along TC pixels
end

% end loop over samples along each row/column for this direction
end

% figure out the overall gray-level gradient for this direction
num_ones = size(find(grad_directional_gradient == 1), 1);
num_neg_ones = size(find(grad_directional_gradient == -1), 1);
if num_ones > num_neg_ones
average_directional_gradient(grad_direction) = 1;
else
average_directional_gradient(grad_direction) = -1;
end

% end loop over possible gradient directions 
end

% figure out overall gray-level gradient for this region
num_ones = size(find(average_directional_gradient == 1), 1);
num_neg_ones = size(find(average_directional_gradient == -1), 1);
if num_ones > num_neg_ones
is_fg_light = 1;
else
is_fg_light = 0;
end

function [new_coordinates new_data] = remove_duplicate_data(cc_coordinates, cc_data, region_data)
%get rid of one of each pair of duplicate connected components
%(duplicate meaning they're in the same position and same size, within small
%error bound),
%eliminating the one from the larger region each time (larger region whose
%edges are not text edges overlaps a smaller region whose edges are text
%edges)

marked_for_removal = [];
next_marked_index = 1;

for outer_index = 1:size(cc_coordinates, 1)
     outer_width = cc_coordinates(outer_index, 4) - cc_coordinates(outer_index, 2);
     outer_height = cc_coordinates(outer_index, 3) - cc_coordinates(outer_index, 1);

for inner_index = outer_index + 1:size(cc_coordinates, 1)

     top_diff = abs(cc_coordinates(outer_index, 1) - cc_coordinates(inner_index, 1));
     left_diff = abs(cc_coordinates(outer_index, 2) - cc_coordinates(inner_index, 2));
     bottom_diff = abs(cc_coordinates(outer_index, 3) - cc_coordinates(inner_index, 3));
     right_diff = abs(cc_coordinates(outer_index, 4) - cc_coordinates(inner_index, 4));

%if all coordinates between outer and inner are close enough, this is a duplicate
if top_diff < outer_height * 0.2 && left_diff < outer_width * 0.2 && bottom_diff < outer_height * 0.2 && right_diff < outer_width * 0.2

str = 'I found a duplicate!'
cc_coordinates(outer_index, :)
cc_coordinates(inner_index, :)

%figure out which regions these two belong to
outer_region = cc_data(outer_index, 1);
inner_region = cc_data(inner_index, 1);

%figure out which region is larger
outer_region_data = region_data(find(region_data(:, 1) == outer_region), :);
inner_region_data = region_data(find(region_data(:, 1) == inner_region), :);
outer_size = outer_region_data(4) * outer_region_data(5);
inner_size = inner_region_data(4) * inner_region_data(5);
if outer_size < inner_size
index_to_remove = inner_index;
else
index_to_remove = outer_index;
end

%if it's not already marked, mark the index of the duplicate to be removed
index_to_remove
find(marked_for_removal == index_to_remove)
size(find(marked_for_removal == index_to_remove))
if size(find(marked_for_removal == index_to_remove), 1) == 0 || size(find(marked_for_removal == index_to_remove), 2) == 0
marked_for_removal(next_marked_index) = index_to_remove
next_marked_index = next_marked_index + 1;
end

%end check for duplicate
end

%end double loop
end
end

str = 'these coordinates will be removed'
cc_coordinates(marked_for_removal,:)

%keep only the cc coordinates and data that do not belong to this index
new_index = 1;
for old_index = 1:size(cc_coordinates, 1)
if size(find(marked_for_removal == old_index), 1) == 0 || size(find(marked_for_removal == old_index), 2) == 0
new_coordinates(new_index, 1:size(cc_coordinates, 2)) = cc_coordinates(old_index, 1:size(cc_coordinates, 2));
new_data(new_index, 1:size(cc_data, 2)) = cc_data(old_index, 1:size(cc_data, 2));
new_index = new_index + 1;
end
end

new_coordinates
new_data
